#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Pre-built deep learning models '''

import os
import warnings

from .Sequential import Sequential
from .blocks import ResBlockBN, ResBlock_Caffe, DenseNetBlock
from .caffe_models import (model_vgg16, model_vgg19, model_resnet50,
                           model_resnet101, model_resnet152)
from .layers import (InputLayer, Conv2d, Pooling, Dense, BN, OutputLayer)
from .model import Model
from .utils import random_name


def LeNet5(conn, model_table='LENET5',
           n_classes=10, n_channels=1, width=28, height=28, scale=1.0 / 255,
           random_flip='none', random_crop='none', offsets=0):
    '''
    Generate a deep learning model with LeNet5 architecture

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, optional
        Specifies the name of CAS table to store the model
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 1
    width : int, optional
        Specifies the width of the input layer
        Default: 28
    height : int, optional
        Specifies the height of the input layer
        Default: 28
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model
        will automatically detect the number of classes based on the
        training set.
        Default: 10
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'none'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data
        is used. Images are cropped to the values that are specified in the
        width and height parameters. Only the images with one or both
        dimensions that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'none'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: 0

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))

    model.add(Conv2d(n_filters=6, width=5, height=5, stride=1))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=16, width=5, height=5, stride=1))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Dense(n=120))
    model.add(Dense(n=84))
    model.add(OutputLayer(n=n_classes))

    return model


def LeNet5_bn(conn, model_table='LENET_BN',
              n_channels=1, width=28, height=28, n_classes=10, scale=1.0 / 255,
              random_flip='none', random_crop='none', offsets=0):
    '''
    Generate a LeNet Model with Batch normalization

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, optional
        Specifies the name of CAS table to store the model
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 1
    width : int, optional
        Specifies the width of the input layer
        Default: 28
    height : int, optional
        Specifies the height of the input layer
        Default: 28
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 10
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1/255
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'none'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'none'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: 0

    Returns
    -------
    :class:`Sequential`
        If `pre_train_weight` is `False`
    :class:`Model`
        If `pre_train_weight` is `True`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop, name='mnist'))

    model.add(Conv2d(n_filters=20, width=5, act='identity', stride=1,
                     includeBias=False, name='conv1'))
    model.add(BN(act='relu', name='conv1_bn'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max', name='pool1'))

    model.add(Conv2d(n_filters=50, width=5, act='identity', stride=1,
                     includeBias=False, name='conv2'))
    model.add(BN(act='relu', name='conv2_bn'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max', name='pool2'))

    model.add(Dense(n=500, name='ip1'))
    model.add(OutputLayer(n=n_classes, name='ip2'))

    return model


def VGG11(conn, model_table='VGG11',
          n_classes=1000, n_channels=3, width=224, height=224, scale=1,
          random_flip='none', random_crop='none', offsets=(103.939, 116.779, 123.68)):
    '''
    Generate a deep learning model with VGG11 architecture

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, optional
        Specifies the name of CAS table to store the model
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'none'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'none'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    if offsets is None:
        offsets = (103.939, 116.779, 123.68)

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))

    model.add(Conv2d(n_filters=64, width=3, height=3, stride=1))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=128, width=3, height=3, stride=1))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1))
    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Dense(n=4096, dropout=0.5))
    model.add(Dense(n=4096, dropout=0.5))
    model.add(OutputLayer(n=n_classes))

    return model


def VGG11_bn(conn, model_table='VGG11',
             n_classes=1000, n_channels=3, width=224, height=224, scale=1,
             random_flip='none', random_crop='none', offsets=(103.939, 116.779, 123.68)):
    '''
    Generate deep learning model with VGG11 architecture with batch normalization layers

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the deep learning model.
        Default: 'VGG11_BN'
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model
        will automatically detect the number of classes based on the
        training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data
        is used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'none'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data
        is used. Images are cropped to the values that are specified in the
        width and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'none'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final input
        data is set after applying scaling and subtracting the specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    if offsets is None:
        offsets = (103.939, 116.779, 123.68)

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))

    model.add(Conv2d(n_filters=64, width=3, height=3, stride=1, act='identity',
                     includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=128, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Dense(n=4096, dropout=0.5))
    model.add(Dense(n=4096, dropout=0.5))
    model.add(OutputLayer(n=n_classes))

    return model


def VGG13(conn, model_table='VGG13',
          n_classes=1000, n_channels=3, width=224, height=224, scale=1,
          random_flip='none', random_crop='none', offsets=(103.939, 116.779, 123.68)):
    '''
    Generate a deep learning model with VGG13 architecture

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, optional
        Specifies the name of CAS table to store the model
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'hv'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default	: 'unique'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    if offsets is None:
        offsets = (103.939, 116.779, 123.68)

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))

    model.add(Conv2d(n_filters=64, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=64, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=128, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=128, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Dense(n=4096, dropout=0.5))
    model.add(Dense(n=4096, dropout=0.5))
    model.add(OutputLayer(n=n_classes))

    return model


def VGG13_bn(conn, model_table='VGG13',
             n_classes=1000, n_channels=3, width=224, height=224, scale=1,
             random_flip='none', random_crop='none', offsets=None):
    '''
    Generate deep learning model with VGG13 architecture with batch normalization layers

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, optional
        Specifies the name of CAS table to store the model
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: None
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'none'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'none'
    offsets : double or iter-of--doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    if offsets is None:
        offsets = (103.939, 116.779, 123.68)

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))

    model.add(Conv2d(n_filters=64, width=3, height=3, stride=1))
    model.add(Conv2d(n_filters=64, width=3, height=3, stride=1))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=128, width=3, height=3, stride=1))
    model.add(Conv2d(n_filters=128, width=3, height=3, stride=1))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1))
    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Dense(n=4096, dropout=0.5))
    model.add(Dense(n=4096, dropout=0.5))
    model.add(OutputLayer(n=n_classes))

    return model


def VGG16(conn, model_table='VGG16',
          n_classes=1000, n_channels=3, width=224, height=224, scale=1,
          random_flip='none', random_crop='none', offsets=(103.939, 116.779, 123.68),
          pre_train_weight=False, pre_train_weight_file=None, include_top=False):
    '''
    Generate a deep learning model with VGG16 architecture

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, optional
        Specifies the name of CAS table to store the model
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'hv'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions that
        are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'unique'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)
    pre_train_weight : boolean, optional
        Specifies whether to use the pre-trained weights from ImageNet data set
        Default: False
    pre_train_weight_file : string, required when pre_train_weight=True.
        Specifies the file name for the pretained weights.
        Must be a fully qualified file name of SAS-compatible file (*.caffemodel.h5)
    include_top : boolean, optional
        Specifies whether to include pre-trained weights of the top layers,
        i.e. the FC layers
        Default: False

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    if not pre_train_weight:
        model = Sequential(conn=conn, model_table=model_table)

        model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                             scale=scale, offsets=offsets, random_flip=random_flip,
                             random_crop=random_crop))

        model.add(Conv2d(n_filters=64, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=64, width=3, height=3, stride=1))
        model.add(Pooling(width=2, height=2, stride=2, pool='max'))

        model.add(Conv2d(n_filters=128, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=128, width=3, height=3, stride=1))
        model.add(Pooling(width=2, height=2, stride=2, pool='max'))

        model.add(Conv2d(n_filters=256, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=256, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=256, width=3, height=3, stride=1))
        model.add(Pooling(width=2, height=2, stride=2, pool='max'))

        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Pooling(width=2, height=2, stride=2, pool='max'))

        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Pooling(width=2, height=2, stride=2, pool='max'))

        model.add(Dense(n=4096, dropout=0.5))
        model.add(Dense(n=4096, dropout=0.5))
        model.add(OutputLayer(n=n_classes))

        return model

    else:
        if pre_train_weight_file is None:
            raise ValueError('\nThe pre-trained weights file is not specified.\n'
                             'Please follow the steps below to attach the pre-trained weights:\n'
                             '1. go to the website https://support.sas.com/documentation/prod-p/vdmml/zip/ '
                             'and download the associated weight file.\n'
                             '2. upload the *.h5 file to '
                             'a server side directory which the CAS session has access to.\n'
                             '3. specify the pre_train_weight_file using the fully qualified server side path.')
        model_cas = model_vgg16.VGG16_Model(
            s=conn, model_table=model_table, n_channels=n_channels,
            width=width, height=height, random_crop=random_crop,
            offsets=offsets)

        if include_top:
            if n_classes != 1000:
                warnings.warn('If include_top = True, n_classes will be set to 1000.', RuntimeWarning)
            model = Model.from_table(model_cas)
            label_table = random_name('label')
            label_file = os.path.join(os.path.dirname(__file__),
                                      'datasources', 'imagenet_label.sas7bdat')
            conn.upload(
                casout=dict(replace=True, name=label_table),
                data=label_file)

            model.load_weights(path=pre_train_weight_file,
                               labeltable=dict(name=label_table,
                                               varS=['levid', 'levname']))
            model._retrieve_('table.droptable', table=label_table)
            return model

        else:
            model = Model.from_table(model_cas, display_note=False)
            model.load_weights(path=pre_train_weight_file)

            weight_table_options = model.model_weights.to_table_params()
            weight_table_options.update(dict(where='_LayerID_<19'))
            model._retrieve_('table.partition', table=weight_table_options,
                             casout=dict(replace=True, **model.model_weights.to_table_params()))
            model._retrieve_('deeplearn.removelayer', model=model_table, name='fc8')
            model._retrieve_('deeplearn.addlayer', model=model_table, name='fc8',
                             layer=dict(type='output', n=n_classes, act='softmax'),
                             srcLayers=['fc7'])
            model = Model.from_table(conn.CASTable(model_table))

            return model


def VGG16_bn(conn, model_table='VGG16',
             n_classes=1000, n_channels=3, width=224, height=224, scale=1,
             random_flip='none', random_crop='none', offsets=(103.939, 116.779, 123.68)):
    '''
    Generate deep learning model with VGG16 architecture with batch normalization layers

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, optional
        Specifies the name of CAS table to store the model
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: None
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'none'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions that
        are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'none'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final input
        data is set after applying scaling and subtracting the specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))

    model.add(Conv2d(n_filters=64, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=64, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=128, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=128, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Dense(n=4096, dropout=0.5))
    model.add(Dense(n=4096, dropout=0.5))
    model.add(OutputLayer(n=n_classes))

    return model


def VGG19(conn, model_table='VGG19',
          n_classes=1000, n_channels=3, width=224, height=224, scale=1,
          random_flip='none', random_crop='none', offsets=(103.939, 116.779, 123.68),
          pre_train_weight=False, pre_train_weight_file=None, include_top=False):
    '''
    Generate a deep learning model with VGG19 architecture

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, optional
        Specifies the name of CAS table to store the model
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'hv'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions that
        are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'unique'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final input
        data is set after applying scaling and subtracting the specified offsets.
        Default: (103.939, 116.779, 123.68)
    pre_train_weight : boolean, optional
        Specifies whether to use the pre-trained weights from ImageNet data set
        Default: False
    pre_train_weight_file : string, required when pre_train_weight=True.
        Specifies the file name for the pretained weights.
        Must be a fully qualified file name of SAS-compatible file (*.caffemodel.h5)
    include_top : boolean, optional
        Specifies whether to include pre-trained weights of the top layers,
        i.e. the FC layers.
        Default: False

    Returns
    -------
    :class:`Sequential`
        If `pre_train_weight` is False
    :class:`Model`
        If `pre_train_teight` is True

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    if not pre_train_weight:
        model = Sequential(conn=conn, model_table=model_table)

        model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                             scale=scale, offsets=offsets, random_flip=random_flip,
                             random_crop=random_crop))

        model.add(Conv2d(n_filters=64, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=64, width=3, height=3, stride=1))
        model.add(Pooling(width=2, height=2, stride=2, pool='max'))

        model.add(Conv2d(n_filters=128, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=128, width=3, height=3, stride=1))
        model.add(Pooling(width=2, height=2, stride=2, pool='max'))

        model.add(Conv2d(n_filters=256, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=256, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=256, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=256, width=3, height=3, stride=1))
        model.add(Pooling(width=2, height=2, stride=2, pool='max'))

        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Pooling(width=2, height=2, stride=2, pool='max'))

        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Conv2d(n_filters=512, width=3, height=3, stride=1))
        model.add(Pooling(width=2, height=2, stride=2, pool='max'))

        model.add(Dense(n=4096, dropout=0.5))
        model.add(Dense(n=4096, dropout=0.5))
        model.add(OutputLayer(n=n_classes))

        return model

    else:
        if pre_train_weight_file is None:
            raise ValueError('\nThe pre-trained weights file is not specified.\n'
                             'Please follow the steps below to attach the pre-trained weights:\n'
                             '1. go to the website https://support.sas.com/documentation/prod-p/vdmml/zip/ '
                             'and download the associated weight file.\n'
                             '2. upload the *.h5 file to '
                             'a server side directory which the CAS session has access to.\n'
                             '3. specify the pre_train_weight_file using the fully qualified server side path.')
        model_cas = model_vgg19.VGG19_Model(
            s=conn, model_table=model_table, n_channels=n_channels,
            width=width, height=height, random_crop=random_crop,
            offsets=offsets)

        if include_top:
            if n_classes != 1000:
                warnings.warn('If include_top = True, n_classes will be set to 1000.', RuntimeWarning)

            model = Model.from_table(model_cas)
            label_table = random_name('label')
            label_file = os.path.join(os.path.dirname(__file__),
                                      'datasources', 'imagenet_label.sas7bdat')
            conn.upload(
                casout=dict(replace=True, name=label_table),
                data=label_file)

            model.load_weights(path=pre_train_weight_file,
                               labeltable=dict(name=label_table,
                                               varS=['levid', 'levname']))
            model._retrieve_('table.droptable', table=label_table)
            return model

        else:

            model = Model.from_table(model_cas, display_note=False)
            model.load_weights(path=pre_train_weight_file)

            weight_table_options = model.model_weights.to_table_params()
            weight_table_options.update(dict(where='_LayerID_<22'))
            model._retrieve_('table.partition', table=weight_table_options,
                             casout=dict(replace=True, **model.model_weights.to_table_params()))
            model._retrieve_('deeplearn.removelayer', model=model_table, name='fc8')
            model._retrieve_('deeplearn.addlayer', model=model_table, name='fc8',
                             layer=dict(type='output', n=n_classes, act='softmax'),
                             srcLayers=['fc7'])
            model = Model.from_table(conn.CASTable(model_table))

            return model


def VGG19_bn(conn, model_table='VGG19',
             n_classes=1000, n_channels=3, width=224, height=224, scale=1,
             random_flip='none', random_crop='none', offsets=(103.939, 116.779, 123.68)):
    '''
    Generate deep learning model with VGG19 architecture with batch normalization layers

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string
        Specifies the name of CAS table to store the model in
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'hv'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions that
        are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default	: 'unique'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final input
        data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))

    model.add(Conv2d(n_filters=64, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=64, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=128, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=128, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=256, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Conv2d(n_filters=512, width=3, height=3, stride=1,
                     act='identity', includeBias=False))
    model.add(BN(act='relu'))
    model.add(Pooling(width=2, height=2, stride=2, pool='max'))

    model.add(Dense(n=4096, dropout=0.5))
    model.add(Dense(n=4096, dropout=0.5))
    model.add(OutputLayer(n=n_classes))

    return model


def ResNet18_SAS(conn, model_table='RESNET18_SAS', batch_norm_first=True,
                 n_classes=1000, n_channels=3, width=224, height=224, scale=1,
                 random_flip='none', random_crop='none',
                 offsets=(103.939, 116.779, 123.68)):
    '''
    Generate a deep learning model with ResNet18 architecture

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, optional
        Specifies the name of CAS table to store the model
        Default: 'RESNET18_SAS'
    batch_norm_first: boolean, optional
        Specifies whether to have batch normalization layer before the
        convolution layer in the residual block.  For a detailed discussion
        about this, please refer to this paper: He, Kaiming, et al. "Identity
        mappings in deep residual networks." European Conference on Computer
        Vision. Springer International Publishing, 2016.
        Default: True
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'none'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions that
        are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default	: 'none'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))
    # Top layers
    model.add(Conv2d(64, 7, act='identity', includeBias=False, stride=2))
    model.add(BN(act='relu'))
    model.add(Pooling(width=3, stride=2))

    kernel_sizes_list = [(3, 3), (3, 3), (3, 3), (3, 3)]
    n_filters_list = [(64, 64), (128, 128), (256, 256), (512, 512)]
    rep_nums_list = [2, 2, 2, 2]

    for i in range(4):
        kernel_sizes = kernel_sizes_list[i]
        n_filters = n_filters_list[i]
        for rep_num in range(rep_nums_list[i]):
            if i == 0:
                strides = 1
            else:
                if rep_num == 0:
                    strides = 2
                else:
                    strides = 1

            model.add(ResBlockBN(kernel_sizes=kernel_sizes, n_filters=n_filters,
                                 strides=strides, batch_norm_first=batch_norm_first))

    # Bottom Layers
    pooling_size = (width // 2 // 2 // 2 // 2 // 2, height // 2 // 2 // 2 // 2 // 2)
    model.add(Pooling(width=pooling_size[0], height=pooling_size[1], pool='mean'))

    model.add(OutputLayer(act='softmax', n=n_classes))

    return model


def ResNet18_Caffe(conn, model_table='RESNET18_CAFFE', batch_norm_first=False,
                   n_classes=1000, n_channels=3, width=224, height=224, scale=1,
                   random_flip='none', random_crop='none', offsets=None):
    '''
    Generate a deep learning model with ResNet18 architecture with convolution shortcut

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the deep learning model.
    batch_norm_first : boolean, optional
        Specifies whether to have batch normalization layer before the
        convolution layer in the residual block.  For a detailed discussion
        about this, please refer to this paper: He, Kaiming, et al. "Identity
        mappings in deep residual networks." European Conference on Computer
        Vision. Springer International Publishing, 2016.
        Default: False
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'hv'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'unique'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))
    # Top layers
    model.add(Conv2d(64, 7, act='identity', includeBias=False, stride=2))
    model.add(BN(act='relu'))
    model.add(Pooling(width=3, stride=2))

    kernel_sizes_list = [(3, 3), (3, 3), (3, 3), (3, 3)]
    n_filters_list = [(64, 64), (128, 128), (256, 256), (512, 512)]
    rep_nums_list = [2, 2, 2, 2]

    for i in range(4):
        kernel_sizes = kernel_sizes_list[i]
        n_filters = n_filters_list[i]
        for rep_num in range(rep_nums_list[i]):
            if rep_num == 0:
                conv_short_cut = True
                if i == 0:
                    strides = 1
                else:
                    strides = 2
            else:
                conv_short_cut = False
                strides = 1
            model.add(ResBlock_Caffe(kernel_sizes=kernel_sizes,
                                     n_filters=n_filters, strides=strides,
                                     batch_norm_first=batch_norm_first,
                                     conv_short_cut=conv_short_cut))

    # Bottom Layers
    pooling_size = (width // 2 // 2 // 2 // 2 // 2, height // 2 // 2 // 2 // 2 // 2)
    model.add(Pooling(width=pooling_size[0], height=pooling_size[1], pool='mean'))

    model.add(OutputLayer(act='softmax', n=n_classes))

    return model


def ResNet34_SAS(conn, model_table='RESNET34_SAS', batch_norm_first=True,
                 n_classes=1000, n_channels=3, width=224, height=224, scale=1,
                 random_flip='none', random_crop='none',
                 offsets=(103.939, 116.779, 123.68)):
    '''
    Generate a deep learning model with ResNet34 architecture

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the deep learning model.
    batch_norm_first: boolean, optional
        Specifies whether to have batch normalization layer before the
        convolution layer in the residual block.  For a detailed discussion
        about this, please refer to this paper: He, Kaiming, et al. "Identity
        mappings in deep residual networks." European Conference on Computer
        Vision. Springer International Publishing, 2016.
        Default: True
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'hv'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'unique'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))
    # Top layers
    model.add(Conv2d(64, 7, act='identity', includeBias=False, stride=2))
    model.add(BN(act='relu'))
    model.add(Pooling(width=3, stride=2))

    kernel_sizes_list = [(3, 3), (3, 3), (3, 3), (3, 3)]
    n_filters_list = [(64, 64), (128, 128), (256, 256), (512, 512)]
    rep_nums_list = [3, 4, 6, 3]

    for i in range(4):
        kernel_sizes = kernel_sizes_list[i]
        n_filters = n_filters_list[i]
        for rep_num in range(rep_nums_list[i]):
            if i == 0:
                strides = 1
            else:
                if rep_num == 0:
                    strides = 2
                else:
                    strides = 1

            model.add(ResBlockBN(kernel_sizes=kernel_sizes, n_filters=n_filters,
                                 strides=strides, batch_norm_first=batch_norm_first))

    # Bottom Layers
    pooling_size = (width // 2 // 2 // 2 // 2 // 2, height // 2 // 2 // 2 // 2 // 2)
    model.add(Pooling(width=pooling_size[0], height=pooling_size[1], pool='mean'))

    model.add(OutputLayer(act='softmax', n=n_classes))

    return model


def ResNet34_Caffe(conn, model_table='RESNET34_CAFFE', batch_norm_first=False,
                   n_classes=1000, n_channels=3, width=224, height=224, scale=1,
                   random_flip='none', random_crop='none', offsets=None):
    '''
    Generate deep learning model with ResNet34 architecture with convolution shortcut

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the deep learning model.
    batch_norm_first : boolean, optional
        Specifies whether to have batch normalization layer before the
        convolution layer in the residual block.  For a detailed discussion
        about this, please refer to this paper: He, Kaiming, et al. "Identity
        mappings in deep residual networks." European Conference on Computer
        Vision. Springer International Publishing, 2016.
        Default: False
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'hv'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions that
        are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'unique'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final input
        data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''

    # conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')
    # model_resnet18.ResNet18_Model(s=conn, model_table=model_table,
    #                               n_classes=n_classes, random_crop=random_crop,
    #                               random_flip=random_flip, offsets=offsets)
    # model = Model.from_table(conn.CASTable(model_table))
    # return model
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))
    # Top layers
    model.add(Conv2d(64, 7, act='identity', includeBias=False, stride=2))
    model.add(BN(act='relu'))
    model.add(Pooling(width=3, stride=2))

    # Configuration of the residual blocks
    kernel_sizes_list = [(3, 3), (3, 3), (3, 3), (3, 3)]
    n_filters_list = [(64, 64), (128, 128), (256, 256), (512, 512)]
    rep_nums_list = [3, 4, 6, 3]

    for i in range(4):
        kernel_sizes = kernel_sizes_list[i]
        n_filters = n_filters_list[i]
        for rep_num in range(rep_nums_list[i]):
            if rep_num == 0:
                conv_short_cut = True
                if i == 0:
                    strides = 1
                else:
                    strides = 2
            else:
                conv_short_cut = False
                strides = 1
            model.add(ResBlock_Caffe(kernel_sizes=kernel_sizes,
                                     n_filters=n_filters, strides=strides,
                                     batch_norm_first=batch_norm_first,
                                     conv_short_cut=conv_short_cut))

    # Bottom Layers
    pooling_size = (width // 2 // 2 // 2 // 2 // 2, height // 2 // 2 // 2 // 2 // 2)
    model.add(Pooling(width=pooling_size[0], height=pooling_size[1], pool='mean'))

    model.add(OutputLayer(act='softmax', n=n_classes))

    return model


def ResNet50_SAS(conn, model_table='RESNET50_SAS', batch_norm_first=True,
                 n_classes=1000, n_channels=3, width=224, height=224, scale=1,
                 random_flip='none', random_crop='none',
                 offsets=(103.939, 116.779, 123.68)):
    '''
    Generate a deep learning model with ResNet50 architecture

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the deep learning model.
    batch_norm_first : boolean, optional
        Specifies whether to have batch normalization layer before the
        convolution layer in the residual block.  For a detailed discussion
        about this, please refer to this paper: He, Kaiming, et al. "Identity
        mappings in deep residual networks." European Conference on Computer
        Vision. Springer International Publishing, 2016.
        Default: True
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'hv'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'unique'
    offsets : double or list-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))
    # Top layers
    model.add(Conv2d(64, 7, act='identity', includeBias=False, stride=2))
    model.add(BN(act='relu'))
    model.add(Pooling(width=3, stride=2))

    kernel_sizes_list = [(1, 3, 1)] * 4
    n_filters_list = [(64, 64, 256), (128, 128, 512), (256, 256, 1024), (512, 512, 2048)]
    rep_nums_list = [3, 4, 6, 3]

    for i in range(4):
        kernel_sizes = kernel_sizes_list[i]
        n_filters = n_filters_list[i]
        for rep_num in range(rep_nums_list[i]):
            if i == 0:
                strides = 1
            else:
                if rep_num == 0:
                    strides = 2
                else:
                    strides = 1

            model.add(ResBlockBN(kernel_sizes=kernel_sizes, n_filters=n_filters,
                                 strides=strides, batch_norm_first=batch_norm_first))

    # Bottom Layers
    pooling_size = (width // 2 // 2 // 2 // 2 // 2, height // 2 // 2 // 2 // 2 // 2)
    model.add(Pooling(width=pooling_size[0], height=pooling_size[1], pool='mean'))

    model.add(OutputLayer(act='softmax', n=n_classes))

    return model


def ResNet50_Caffe(conn, model_table='RESNET50_CAFFE', batch_norm_first=False,
                   n_classes=1000, n_channels=3, width=224, height=224, scale=1,
                   random_flip='none', random_crop='none',
                   offsets=(103.939, 116.779, 123.68),
                   pre_train_weight=False, pre_train_weight_file=None, include_top=False):
    '''
    Generate deep learning model with ResNet50 architecture with convolution shortcut

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the deep learning model.
    batch_norm_first : boolean, optional
        Specifies whether to have batch normalization layer before the
        convolution layer in the residual block.  For a detailed discussion
        about this, please refer to this paper: He, Kaiming, et al. "Identity
        mappings in deep residual networks." European Conference on Computer
        Vision. Springer International Publishing, 2016.
        Default: False
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'none'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'none'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)
    pre_train_weight : boolean, optional
        Specifies whether to use the pre-trained weights from ImageNet data set
        Default: False
    pre_train_weight_file : string, required when pre_train_weight=True.
        Specifies the file name for the pretained weights.
        Must be a fully qualified file name of SAS-compatible file (*.caffemodel.h5)
    include_top : boolean, optional
        Specifies whether to include pre-trained weights of the top layers,
        i.e. the last layer for classification.
        Default: False

    Returns
    -------
    :class:`Sequential`
        If `pre_train_weight` is `False`
    :class:`Model`
        If `pre_train_weight` is `True`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    if not pre_train_weight:
        model = Sequential(conn=conn, model_table=model_table)

        model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                             scale=scale, offsets=offsets, random_flip=random_flip,
                             random_crop=random_crop))
        # Top layers
        model.add(Conv2d(64, 7, act='identity', includeBias=False, stride=2))
        model.add(BN(act='relu'))
        model.add(Pooling(width=3, stride=2))
        # Residual block configuration.
        kernel_sizes_list = [(1, 3, 1)] * 4
        n_filters_list = [(64, 64, 256), (128, 128, 512),
                          (256, 256, 1024), (512, 512, 2048)]
        rep_nums_list = [3, 4, 6, 3]

        for i in range(4):
            kernel_sizes = kernel_sizes_list[i]
            n_filters = n_filters_list[i]
            for rep_num in range(rep_nums_list[i]):
                if rep_num == 0:
                    conv_short_cut = True
                    if i == 0:
                        strides = 1
                    else:
                        strides = 2
                else:
                    conv_short_cut = False
                    strides = 1
                model.add(ResBlock_Caffe(kernel_sizes=kernel_sizes,
                                         n_filters=n_filters, strides=strides,
                                         batch_norm_first=batch_norm_first,
                                         conv_short_cut=conv_short_cut))

        # Bottom Layers
        pooling_size = (width // 2 // 2 // 2 // 2 // 2, height // 2 // 2 // 2 // 2 // 2)
        model.add(Pooling(width=pooling_size[0], height=pooling_size[1], pool='mean'))

        model.add(OutputLayer(act='softmax', n=n_classes))

        return model

    else:
        if pre_train_weight_file is None:
            raise ValueError('\nThe pre-trained weights file is not specified.\n'
                             'Please follow the steps below to attach the pre-trained weights:\n'
                             '1. go to the website https://support.sas.com/documentation/prod-p/vdmml/zip/ '
                             'and download the associated weight file.\n'
                             '2. upload the *.h5 file to '
                             'a server side directory which the CAS session has access to.\n'
                             '3. specify the pre_train_weight_file using the fully qualified server side path.')
        model_cas = model_resnet50.ResNet50_Model(
            s=conn, model_table=model_table, n_channels=n_channels,
            width=width, height=height, random_crop=random_crop,
            offsets=offsets)

        if include_top:
            if n_classes != 1000:
                warnings.warn('If include_top = True, n_classes will be set to 1000.', RuntimeWarning)

            model = Model.from_table(model_cas)
            label_table = random_name('label')
            label_file = os.path.join(os.path.dirname(__file__),
                                      'datasources', 'imagenet_label.sas7bdat')
            conn.upload(
                casout=dict(replace=True, name=label_table),
                data=label_file)

            model.load_weights(path=pre_train_weight_file,
                               labeltable=dict(name=label_table,
                                               varS=['levid', 'levname']))
            model._retrieve_('table.droptable', table=label_table)
            return model

        else:
            model = Model.from_table(model_cas, display_note=False)
            model.load_weights(path=pre_train_weight_file)
            model._retrieve_('deeplearn.removelayer', model=model_table, name='fc1000')
            model._retrieve_('deeplearn.addlayer', model=model_table, name='output',
                             layer=dict(type='output', n=n_classes, act='softmax'),
                             srcLayers=['pool5'])

            weight_table_options = model.model_weights.to_table_params()
            weight_table_options.update(dict(where='_LayerID_<125'))
            model._retrieve_('table.partition', table=weight_table_options,
                             casout=dict(replace=True, **model.model_weights.to_table_params()))
            model = Model.from_table(conn.CASTable(model_table))
            return model


def ResNet101_SAS(conn, model_table='RESNET101_SAS', batch_norm_first=True,
                  n_classes=1000, n_channels=3, width=224, height=224, scale=1,
                  random_flip='none', random_crop='none',
                  offsets=(103.939, 116.779, 123.68)):
    '''
    Generate a deep learning model with ResNet101 architecture

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the deep learning model.
    batch_norm_first : boolean, optional
        Specifies whether to have batch normalization layer before the
        convolution layer in the residual block.  For a detailed discussion
        about this, please refer to this paper: He, Kaiming, et al. "Identity
        mappings in deep residual networks." European Conference on Computer
        Vision. Springer International Publishing, 2016.
        Default: True
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data
        is used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'hv'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default	: 'unique'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final input
        data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))
    # Top layers
    model.add(Conv2d(64, 7, act='identity', includeBias=False, stride=2))
    model.add(BN(act='relu'))
    model.add(Pooling(width=3, stride=2))

    kernel_sizes_list = [(1, 3, 1)] * 4
    n_filters_list = [(64, 64, 256), (128, 128, 512), (256, 256, 1024), (512, 512, 2048)]
    rep_nums_list = [3, 4, 23, 3]

    for i in range(4):
        kernel_sizes = kernel_sizes_list[i]
        n_filters = n_filters_list[i]
        for rep_num in range(rep_nums_list[i]):
            if i == 0:
                strides = 1
            else:
                if rep_num == 0:
                    strides = 2
                else:
                    strides = 1

            model.add(ResBlockBN(kernel_sizes=kernel_sizes, n_filters=n_filters,
                                 strides=strides, batch_norm_first=batch_norm_first))

    # Bottom Layers
    pooling_size = (width // 2 // 2 // 2 // 2 // 2, height // 2 // 2 // 2 // 2 // 2)
    model.add(Pooling(width=pooling_size[0], height=pooling_size[1], pool='mean'))

    model.add(OutputLayer(act='softmax', n=n_classes))

    return model


def ResNet101_Caffe(conn, model_table='RESNET101_CAFFE', batch_norm_first=False,
                    n_classes=1000, n_channels=3, width=224, height=224, scale=1,
                    random_flip='none', random_crop='none',
                    offsets=(103.939, 116.779, 123.68),
                    pre_train_weight=False, pre_train_weight_file=None, include_top=False):
    '''
    Generate deep learning model with ResNet101 architecture with convolution shortcut

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the deep learning model.
    batch_norm_first : boolean, optional
        Specifies whether to have batch normalization layer before the
        convolution layer in the residual block.  For a detailed discussion
        about this, please refer to this paper: He, Kaiming, et al. "Identity
        mappings in deep residual networks." European Conference on Computer
        Vision. Springer International Publishing, 2016.
        Default: False
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'none'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'none'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)
    pre_train_weight : boolean, optional
        Specifies whether to use the pre-trained weights from ImageNet data set
        Default: False
    pre_train_weight_file : string, required when pre_train_weight=True.
        Specifies the file name for the pretained weights.
        Must be a fully qualified file name of SAS-compatible file (*.caffemodel.h5)
    include_top : boolean, optional
        Specifies whether to include pre-trained weights of the top layers,
        i.e. the last layer for classification.
        Default: False.

    Returns
    -------
    :class:`Sequential`
        If `pre_train_weight` is `False`
    :class:`Model`
        If `pre_train_weight` is `True`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    if not pre_train_weight:
        model = Sequential(conn=conn, model_table=model_table)

        model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                             scale=scale, offsets=offsets, random_flip=random_flip,
                             random_crop=random_crop))
        # Top layers
        model.add(Conv2d(64, 7, act='identity', includeBias=False, stride=2))
        model.add(BN(act='relu'))
        model.add(Pooling(width=3, stride=2))
        # Residual block configuration.
        kernel_sizes_list = [(1, 3, 1)] * 4
        n_filters_list = [(64, 64, 256), (128, 128, 512),
                          (256, 256, 1024), (512, 512, 2048)]
        rep_nums_list = [3, 4, 23, 3]

        for i in range(4):
            kernel_sizes = kernel_sizes_list[i]
            n_filters = n_filters_list[i]
            for rep_num in range(rep_nums_list[i]):
                if rep_num == 0:
                    conv_short_cut = True
                    if i == 0:
                        strides = 1
                    else:
                        strides = 2
                else:
                    conv_short_cut = False
                    strides = 1
                model.add(ResBlock_Caffe(kernel_sizes=kernel_sizes,
                                         n_filters=n_filters, strides=strides,
                                         batch_norm_first=batch_norm_first,
                                         conv_short_cut=conv_short_cut))

        # Bottom Layers
        pooling_size = (width // 2 // 2 // 2 // 2 // 2, height // 2 // 2 // 2 // 2 // 2)
        model.add(Pooling(width=pooling_size[0], height=pooling_size[1], pool='mean'))

        model.add(OutputLayer(act='softmax', n=n_classes))

        return model

    else:
        if pre_train_weight_file is None:
            raise ValueError('\nThe pre-trained weights file is not specified.\n'
                             'Please follow the steps below to attach the pre-trained weights:\n'
                             '1. go to the website https://support.sas.com/documentation/prod-p/vdmml/zip/ '
                             'and download the associated weight file.\n'
                             '2. upload the *.h5 file to '
                             'a server side directory which the CAS session has access to.\n'
                             '3. specify the pre_train_weight_file using the fully qualified server side path.')
        model_cas = model_resnet101.ResNet101_Model(
            s=conn, model_table=model_table, n_channels=n_channels,
            width=width, height=height, random_crop=random_crop,
            offsets=offsets)

        if include_top:
            if n_classes != 1000:
                warnings.warn('If include_top = True, n_classes will be set to 1000.', RuntimeWarning)

            model = Model.from_table(model_cas)
            label_table = random_name('label')
            label_file = os.path.join(os.path.dirname(__file__),
                                      'datasources', 'imagenet_label.sas7bdat')
            conn.upload(
                casout=dict(replace=True, name=label_table),
                data=label_file)

            model.load_weights(path=pre_train_weight_file,
                               labeltable=dict(name=label_table,
                                               varS=['levid', 'levname']))
            model._retrieve_('table.droptable', table=label_table)
            return model

        else:
            model = Model.from_table(conn.CASTable(model_table), display_note=False)
            model.load_weights(path=pre_train_weight_file)
            model._retrieve_('deeplearn.removelayer', model=model_table, name='fc1000')
            model._retrieve_('deeplearn.addlayer', model=model_table, name='output',
                             layer=dict(type='output', n=n_classes, act='softmax'),
                             srcLayers=['pool5'])

            weight_table_options = model.model_weights.to_table_params()
            weight_table_options.update(dict(where='_LayerID_<244'))
            model._retrieve_('table.partition', table=weight_table_options,
                             casout=dict(replace=True, **model.model_weights.to_table_params()))
            model = Model.from_table(conn.CASTable(model_table))
            return model


def ResNet152_SAS(conn, model_table='RESNET152_SAS', batch_norm_first=True,
                  n_classes=1000, n_channels=3, width=224, height=224, scale=1,
                  random_flip='none', random_crop='none',
                  offsets=(103.939, 116.779, 123.68)):
    '''
    Generate a deep learning model with ResNet152 architecture

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the deep learning model.
    batch_norm_first : boolean, optional
        Specifies whether to have batch normalization layer before the
        convolution layer in the residual block.  For a detailed discussion
        about this, please refer to this paper: He, Kaiming, et al. "Identity
        mappings in deep residual networks." European Conference on Computer
        Vision. Springer International Publishing, 2016.
        Default: True
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'hv'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'unique'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))
    # Top layers
    model.add(Conv2d(64, 7, act='identity', includeBias=False, stride=2))
    model.add(BN(act='relu'))
    model.add(Pooling(width=3, stride=2))

    kernel_sizes_list = [(1, 3, 1)] * 4
    n_filters_list = [(64, 64, 256), (128, 128, 512), (256, 256, 1024), (512, 512, 2048)]
    rep_nums_list = [3, 8, 36, 3]

    for i in range(4):
        kernel_sizes = kernel_sizes_list[i]
        n_filters = n_filters_list[i]
        for rep_num in range(rep_nums_list[i]):
            if i == 0:
                strides = 1
            else:
                if rep_num == 0:
                    strides = 2
                else:
                    strides = 1

            model.add(ResBlockBN(kernel_sizes=kernel_sizes, n_filters=n_filters,
                                 strides=strides, batch_norm_first=batch_norm_first))

    # Bottom Layers
    pooling_size = (width // 2 // 2 // 2 // 2 // 2,
                    height // 2 // 2 // 2 // 2 // 2)
    model.add(Pooling(width=pooling_size[0], height=pooling_size[1], pool='mean'))

    model.add(OutputLayer(act='softmax', n=n_classes))

    return model


def ResNet152_Caffe(conn, model_table='RESNET152_CAFFE', batch_norm_first=False,
                    n_classes=1000, n_channels=3, width=224, height=224, scale=1,
                    random_flip='none', random_crop='none',
                    offsets=(103.939, 116.779, 123.68),
                    pre_train_weight=False, pre_train_weight_file=None, include_top=False):
    '''
    Generate deep learning model with ResNet152 architecture with convolution shortcut

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the deep learning model.
    batch_norm_first : boolean, optional
        Specifies whether to have batch normalization layer before the
        convolution layer in the residual block.  For a detailed discussion
        about this, please refer to this paper: He, Kaiming, et al. "Identity
        mappings in deep residual networks." European Conference on Computer
        Vision. Springer International Publishing, 2016.
        Default: False
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'none'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'none'
    offsets : double, or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)
    pre_train_weight : boolean, optional
        Specifies whether to use the pre-trained weights from ImageNet data set
        Default: False
    pre_train_weight_file : string, required when pre_train_weight=True.
        Specifies the file name for the pretained weights.
        Must be a fully qualified file name of SAS-compatible file (*.caffemodel.h5)
    include_top : boolean, optional
        Specifies whether to include pre-trained weights of the top layers,
        i.e. the last layer for classification.
        Default : False

    Returns
    -------
    :class:`Sequential`
        If `pre_train_weight` is `False`
    :class:`Model`
        If `pre_train_weight` is `True`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    if not pre_train_weight:
        model = Sequential(conn=conn, model_table=model_table)

        model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                             scale=scale, offsets=offsets, random_flip=random_flip,
                             random_crop=random_crop))
        # Top layers
        model.add(Conv2d(64, 7, act='identity', includeBias=False, stride=2))
        model.add(BN(act='relu'))
        model.add(Pooling(width=3, stride=2))
        # Residual block configuration.
        kernel_sizes_list = [(1, 3, 1)] * 4
        n_filters_list = [(64, 64, 256), (128, 128, 512),
                          (256, 256, 1024), (512, 512, 2048)]
        rep_nums_list = [3, 8, 36, 3]

        for i in range(4):
            kernel_sizes = kernel_sizes_list[i]
            n_filters = n_filters_list[i]
            for rep_num in range(rep_nums_list[i]):
                if rep_num == 0:
                    conv_short_cut = True
                    if i == 0:
                        strides = 1
                    else:
                        strides = 2
                else:
                    conv_short_cut = False
                    strides = 1
                model.add(ResBlock_Caffe(kernel_sizes=kernel_sizes,
                                         n_filters=n_filters, strides=strides,
                                         batch_norm_first=batch_norm_first,
                                         conv_short_cut=conv_short_cut))

        # Bottom Layers
        pooling_size = (width // 2 // 2 // 2 // 2 // 2,
                        height // 2 // 2 // 2 // 2 // 2)
        model.add(Pooling(width=pooling_size[0], height=pooling_size[1], pool='mean'))

        model.add(OutputLayer(act='softmax', n=n_classes))

        return model
    else:
        if pre_train_weight_file is None:
            raise ValueError('\nThe pre-trained weights file is not specified.\n'
                             'Please follow the steps below to attach the pre-trained weights:\n'
                             '1. go to the website https://support.sas.com/documentation/prod-p/vdmml/zip/ '
                             'and download the associated weight file.\n'
                             '2. upload the *.h5 file to '
                             'a server side directory which the CAS session has access to.\n'
                             '3. specify the pre_train_weight_file using the fully qualified server side path.')
        model_cas = model_resnet152.ResNet152_Model(
            s=conn, model_table=model_table, n_channels=n_channels,
            width=width, height=height, random_crop=random_crop,
            offsets=offsets)

        if include_top:
            if n_classes != 1000:
                warnings.warn('If include_top = True, n_classes will be set to 1000.', RuntimeWarning)

            model = Model.from_table(model_cas)
            label_table = random_name('label')
            label_file = os.path.join(os.path.dirname(__file__),
                                      'datasources', 'imagenet_label.sas7bdat')
            conn.upload(
                casout=dict(replace=True, name=label_table),
                data=label_file)

            model.load_weights(path=pre_train_weight_file,
                               labeltable=dict(name=label_table,
                                               varS=['levid', 'levname']))
            model._retrieve_('table.droptable', table=label_table)
            return model

        else:
            model = Model.from_table(conn.CASTable(model_table), display_note=False)
            model.load_weights(path=pre_train_weight_file)
            model._retrieve_('deeplearn.removelayer', model=model_table, name='fc1000')
            model._retrieve_('deeplearn.addlayer', model=model_table, name='output',
                             layer=dict(type='output', n=n_classes, act='softmax'),
                             srcLayers=['pool5'])

            weight_table_options = model.model_weights.to_table_params()
            weight_table_options.update(dict(where='_LayerID_<363'))
            model._retrieve_('table.partition', table=weight_table_options,
                             casout=dict(replace=True, **model.model_weights.to_table_params()))
            model = Model.from_table(conn.CASTable(model_table))
            return model


def wide_resnet(conn, model_table='WIDE_RESNET', batch_norm_first=True, depth=2,
                k=4, n_classes=None, n_channels=3, width=32, height=32, scale=1,
                random_flip='none', random_crop='none', offsets=(114, 122, 125)):
    '''
    Generate a deep learning model with ResNet152 architecture

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the deep learning model.
    batch_norm_first : boolean, optional
        Specifies whether to have batch normalization layer before the
        convolution layer in the residual block.  For a detailed discussion
        about this, please refer to this paper: He, Kaiming, et al. "Identity
        mappings in deep residual networks." European Conference on Computer
        Vision. Springer International Publishing, 2016.
        Default: True
    depth : int
        Specifies the number of convolution layers added into the model
        Default: 2
    k : int
        Specifies the widening factor
        Default: 4
    n_classes : int, optional
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: 1000
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    scale : double, optional
        Specifies a scaling factor to apply to each image
        Default: 1
    random_flip : string, optional
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Valid Values: 'h', 'hv', 'none', 'v'
        Default: 'hv'
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default: 'unique'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    :class:`Sequential`

    '''
    conn.retrieve('loadactionset', _messagelevel='error', actionset='deeplearn')

    n_stack = int((depth - 2) / 6)
    in_filters = 16

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))
    # Top layers
    model.add(Conv2d(in_filters, 3, act='identity', includeBias=False, stride=1))
    model.add(BN(act='relu'))

    # Residual block configuration.
    n_filters_list = [(16 * k, 16 * k), (32 * k, 32 * k), (64 * k, 64 * k)]
    kernel_sizes_list = [(3, 3)] * len(n_filters_list)
    rep_nums_list = [n_stack, n_stack, n_stack]

    for i in range(len(n_filters_list)):
        kernel_sizes = kernel_sizes_list[i]
        n_filters = n_filters_list[i]
        for rep_num in range(rep_nums_list[i]):
            if i == 0:
                strides = 1
            else:
                if rep_num == 0:
                    strides = 2
                else:
                    strides = 1

            model.add(ResBlockBN(kernel_sizes=kernel_sizes, n_filters=n_filters,
                                 strides=strides, batch_norm_first=batch_norm_first))
    model.add(BN(act='relu'))
    # Bottom Layers
    pooling_size = (width // 2 // 2, height // 2 // 2)
    model.add(Pooling(width=pooling_size[0], height=pooling_size[1], pool='mean'))

    model.add(OutputLayer(act='softmax', n=n_classes))

    return model


def DenseNet_Cifar(conn, model_table='DenseNet_Cifar', n_classes=None, conv_channel=16, growth_rate=12,
                   n_blocks=4, n_cells=4, n_channels=3, width=32, height=32, scale=1,
                   random_flip='none', random_crop='none', offsets=(85, 111, 139)):
    '''
    Generate a deep learning model with DenseNet architecture

    Parameters
    ----------
    conn :
        Specifies the connection of the CAS connection.
    model_table : string
        Specifies the name of CAS table to store the model.
    n_classes : int, optional.
        Specifies the number of classes. If None is assigned, the model will
        automatically detect the number of classes based on the training set.
        Default: None
    conv_channel: int, optional.
        Specifies the number of filters of first convolution layer.
        Default : 16
    growth_rate: int, optional.
        Specifies growth rate of convolution layer.
        Default : 12
    n_blocks : int, optional.
        Specifies the number of DenseNetBlocks.
        Default : 4
    n_cells : int, optional.
        Specifies the number of densely connection in each DenseNetBlock
        Default : 4
    n_channels : double, optional.
        Specifies the number of the channels of the input layer.
        Default : 3.
    width : double, optional.
        Specifies the width of the input layer.
        Default : 224.
    height : double, optional.
        Specifies the height of the input layer.
        Default : 224.
    scale : double, optional.
        Specifies a scaling factor to apply to each image..
        Default : 1.
    random_flip : string, "h" | "hv" | "none" | "v"
        Specifies how to flip the data in the input layer when image data is
        used. Approximately half of the input data is subject to flipping.
        Default	: "hv"
    random_crop : string, "none" or "unique"
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters. Only the images with one or both dimensions that
        are larger than those sizes are cropped.
        Default	: "unique"
    offsets : (double-1 <, double-2, ...>), optional
        Specifies an offset for each channel in the input data. The final input
        data is set after applying scaling and subtracting the specified offsets.
        Default : (85, 111, 139)

    Returns
    -------
    :class:`Sequential`
        A model object using DenseNet_Cifar architecture.

    '''

    channel_in = conv_channel  # number of channel of transition conv layer

    model = Sequential(conn=conn, model_table=model_table)

    model.add(InputLayer(n_channels=n_channels, width=width, height=height,
                         scale=scale, offsets=offsets, random_flip=random_flip,
                         random_crop=random_crop))
    # Top layers
    model.add(Conv2d(conv_channel, width=3, act='identity', includeBias=False, stride=1))

    for i in range(n_blocks):
        model.add(DenseNetBlock(n_cells=n_cells, kernel_size=3,
                                n_filter=growth_rate, stride=1))
        # transition block
        channel_in += (growth_rate * n_cells)
        model.add(BN(act='relu'))
        if i != (n_blocks - 1):
            model.add(Conv2d(channel_in, width=3, act='identity',
                             includeBias=False, stride=1))
            model.add(Pooling(width=2, height=2, pool='mean'))

    # Bottom Layers
    pooling_size = (width // (2 ** n_blocks // 2), height // (2 ** n_blocks // 2))
    model.add(Pooling(width=pooling_size[0], height=pooling_size[1], pool='mean'))

    model.add(OutputLayer(act='softmax', n=n_classes))

    return model



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from .sas_keras_parse import keras_to_sas
# from .sas_caffe_parse import caffe_to_sas



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Special functionality for CAS tables containing image data '''

import matplotlib.pyplot as plt
import numpy as np
from swat.cas.table import CASTable

from .utils import random_name, image_blocksize


class ImageTable(CASTable):
    '''
    Specialized CASTable for Image Data

    Parameters
    ----------
    name : string
        The name of the CAS table
    **table_params : keyword arguments, optional
        Parameters to the :class:`CASTable` constructor

    Attributes
    ----------
    image_summary : pandas.Series
        The summary of the images contained in the image table.
    label_freq : pandas.Series
        The count of images in different categories.
    channel_means : tuple of double
        The mean of the image intensities in each channels.


    Returns
    -------
    :class:`ImageTable`

    '''

    def __init__(self, name, **table_params):
        CASTable.__init__(self, name, **table_params)
        self.patch_level = 0

    @classmethod
    def from_table(cls, tbl, image_col='_image_', label_col='_label_',
                   path_col=None, columns=None, casout=None):
        '''
        Create an ImageTable from a CASTable

        Parameters
        ----------
        tbl : CASTable
            The CASTable object to use as the source.
        image_col : str, optional
            Specifies the column name for the image data.
            Default = '_image_'
        label_col : str, optional
            Specifies the column name for the labels.
            Default = '_label_'
        path_col : str, optional
            Specifies the column name that stores the path for each image.
            Default = None, and the unique image ID will be generated from the labels.
        columns : list of str, optional
            Specifies the extra columns in the image table.
            Default = None
        casout : dict
            Specifies the output CASTable parameters.
            Default = None.
            Note : the options of replace=True, blocksize=32 will be automatically
            added to the casout option.

        Returns
        -------
        :class:`ImageTable`

        '''
        out = cls(**tbl.params)

        conn = tbl.get_connection()
        conn.loadactionset('image', _messagelevel='error')

        if casout is None:
            casout = {}
        elif isinstance(casout, CASTable):
            casout = casout.to_outtable_params()

        if 'name' not in casout:
            casout['name'] = random_name()

        if '_filename_0' in tbl.columninfo().ColumnInfo.Column.tolist():
            computedvars = []
            code = []
        else:
            computedvars = ['_filename_0']
            code = ['length _filename_0 varchar(*);']
            if path_col is not None:
                code.append(('_loc1 = LENGTH({0}) - '
                             'INDEX(REVERSE({0}),\'/\')+2;').format(path_col))
                code.append('_filename_0 = SUBSTR({},_loc1);'.format(path_col))
            else:
                code.append('call streaminit(-1);shuffle_id=rand("UNIFORM")*10**10;')
                code.append(('_filename_0=cats({},"_",put(put(shuffle_id,z10.)'
                             ',$char10.),".jpg");').format(label_col))

        if image_col != '_image_':
            computedvars.append('_image_')
            code.append('_image_ = {};'.format(image_col))

        if label_col != '_label_':
            computedvars.append('_label_')
            code.append('_label_ = {};'.format(label_col))

        code = '\n'.join(code)

        if computedvars:
            table_opts = dict(computedvars=computedvars,
                              computedvarsprogram=code,
                              **tbl.params)
        else:
            table_opts = dict(**tbl.params)

        # This will generate the '_image_' and '_label_' columns.
        conn.retrieve('table.shuffle', _messagelevel='error',
                      table=table_opts,
                      casout=dict(replace=True, blocksize=32, **casout))

        column_names = ['_image_', '_label_', '_filename_0']
        if columns is not None:
            if not isinstance(columns, list):
                columns = list(columns)
            column_names += columns

        # Remove the unwanted columns.
        conn.retrieve('table.partition', _messagelevel='error',
                      table=dict(Vars=column_names, **casout),
                      casout=dict(replace=True, blocksize=32, **casout))

        out = cls(**casout)
        out.set_connection(conn)

        return out

    @classmethod
    def load_files(cls, conn, path, casout=None, columns=None, **kwargs):
        '''
        Create ImageTable from files in `path`

        Parameters
        ----------
        conn : CAS
            The CAS connection object
        path : string
            The path to the image directory on the server
        casout : dict, optional
            The output table specifications
        columns : list of str, optional
            Specifies the extra columns in the image table.
        **kwargs : keyword arguments, optional
            Additional keyword arguments to the `image.loadimages` action

        Returns
        -------
        :class:`ImageTable`

        '''
        conn.loadactionset('image', _messagelevel='error')

        if casout is None:
            casout = dict(name=random_name())
        elif isinstance(casout, CASTable):
            casout = casout.to_outtable_params()

        if 'name' not in casout:
            casout['name'] = random_name()

        conn.retrieve('image.loadimages', _messagelevel='error',
                      casout=casout,
                      distribution=dict(type='random'),
                      recurse=True, labellevels=-1,
                      path=path, **kwargs)

        code = []
        code.append('length _filename_0 varchar(*);')
        code.append('_loc1 = LENGTH(_path_) - INDEX(REVERSE(_path_),\'/\')+2;')
        code.append('_filename_0 = SUBSTR(_path_,_loc1);')
        code = '\n'.join(code)
        column_names = ['_image_', '_label_', '_filename_0']
        if columns is not None:
            column_names += columns
        conn.retrieve('table.partition', _messagelevel='error',
                      table=dict(Vars=column_names,
                                 computedvars=['_filename_0'],
                                 computedvarsprogram=code,
                                 **casout),
                      casout=dict(replace=True, blocksize=32, **casout))

        out = cls(**casout)
        out.set_connection(conn)
        return out

    def __copy__(self):
        out = CASTable.__copy__(self)
        out.patch_level = self.patch_level
        return out

    def __deepcopy__(self, memo):
        out = CASTable.__deepcopy__(self, memo)
        out.patch_level = self.patch_level
        return out

    def to_files(self, path):
        '''
        Save the images in the original format under the specified directory

        Parameters
        ----------
        path : string
            Specifies the directory on the server to save the images

        '''
        caslib = random_name('Caslib', 6)
        self._retrieve('addcaslib', name=caslib, path=path, activeonadd=False)

        file_name = '_filename_{}'.format(self.patch_level)
        self._retrieve('image.saveimages', caslib=caslib,
                       images=dict(table=self.to_table_params(), path=file_name),
                       labellevels=1)

        self._retrieve('dropcaslib', caslib=caslib)

    def to_sashdat(self, path=None, name=None, **kwargs):
        '''
        Save the ImageTable to a sashdat file

        Parameters
        ----------
        path : string
            Specifies the directory on the server to save the images

        '''
        caslib = random_name('Caslib', 6)
        self._retrieve('addcaslib', name=caslib, path=path, activeonadd=False,
                       datasource=dict(srcType='DNFS'))
        if name is None:
            name = self.to_params()['name'] + '.sashdat'

        self._retrieve('table.save', caslib=caslib, name=name,
                       table=self.to_params(), **kwargs)
        self._retrieve('dropcaslib', caslib=caslib)

    def copy_table(self, casout=None):
        '''
        Create a copy of the ImageTable

        Parameters
        ----------
        casout : dict, optional
            Output CAS table parameters

        Returns
        -------
        :class:`ImageTable`

        '''
        if casout is None:
            casout = {}
            casout['name'] = random_name()

        res = self._retrieve('table.partition', casout=casout, table=self)['casTable']
        out = ImageTable.from_table(tbl=res)
        out.params.update(res.params)

        return out

    def show(self, nimages=5, ncol=8, randomize=False, figsize=None):
        '''
        Display a grid of images

        Parameters
        ----------
        nimages : int, optional
            Specifies the number of images to be displayed.
            If nimage is greater than the maximum number of images in the
            table, it will be set to this maximum number.
            Note: Specifying a large value for nimages can lead to slow
            performance.
        ncol : int, optional
            Specifies the layout of the display, determine the number of
            columns in the plots.
        randomize : boolean, optional
            Specifies whether to randomly choose the images for display.

        '''
        nimages = min(nimages, len(self))

        if randomize:
            temp_tbl = self.retrieve('image.fetchimages', _messagelevel='error',
                                     table=dict(
                                         computedvars=['random_index'],
                                         computedvarsprogram='call streaminit(-1);'
                                                             'random_index='
                                                             'rand("UNIFORM");',
                                         **self.to_table_params()),
                                     sortby='random_index', to=nimages)
        else:
            temp_tbl = self._retrieve('image.fetchimages', to=nimages)

        if nimages > ncol:
            nrow = nimages // ncol + 1
        else:
            nrow = 1
            ncol = nimages
        if figsize is None:
            figsize = (16, 16 // ncol * nrow)
        fig = plt.figure(figsize=figsize)

        for i in range(nimages):
            image = temp_tbl['Images']['Image'][i]
            label = temp_tbl['Images']['Label'][i]
            ax = fig.add_subplot(nrow, ncol, i + 1)
            ax.set_title('{}'.format(label))
            if len(image.size) == 2:
                plt.imshow(np.array(image), cmap='Greys_r')
            else:
                plt.imshow(image)
            plt.xticks([]), plt.yticks([])

    def crop(self, x=0, y=0, width=None, height=None, inplace=True):
        '''
        Crop the images in the ImageTable

        Parameters
        ----------
        x : int, optional
            Specify the x location of the top-left corner of the cropped images.
        y : int, optional
            Specify the y location of the top-left corner of the cropped images.
        width : int, optional
            Specify the width of the cropped images.
        height : int, optional
            Specify the height of the cropped images.
            If not specified, height will be set to be equal to width.
        inplace: boolean, optional
            Specifies whether to update the original table, or to create a new one.

        Returns
        -------
        :class:`ImageTable`
            If `inplace=False`
        `None`
            If `inplace=True`

        '''
        if (width is None) and (height is None):
            width = 224
        if width is None:
            width = height
        if height is None:
            height = width
        blocksize = image_blocksize(width, height)

        column_names = ['_filename_{}'.format(i) for i in range(self.patch_level + 1)]

        if inplace:
            self._retrieve('image.processimages',
                           copyvars=column_names,
                           casout=dict(replace=True, blocksize=blocksize,
                                       **self.to_outtable_params()),
                           imagefunctions=[
                               dict(functionoptions=dict(functiontype='GET_PATCH',
                                                         x=x, y=y,
                                                         w=width, h=height))])

        else:
            out = self.copy_table()
            out.crop(x=x, y=x, width=width, height=height)
            return out

    def resize(self, width=None, height=None, inplace=True):
        '''
        Resize the images in the ImageTable

        Parameters
        ----------
        width : int, optional
            Specify the target width of the resized images.
        height : int, optional
            Specify the target height of the resized images.
            If not specified, height will be set to be equal to width.
        inplace: boolean, optional
            Specifies whether to update the original table, or to create
            a new one.

        Returns
        -------
        :class:`ImageTable`
            If `inplace=False`
        `None`
            If `inplace=True`

        '''
        if (width is None) and (height is None):
            width = 224
        if width is None:
            width = height
        if height is None:
            height = width
        blocksize = image_blocksize(width, height)
        column_names = ['_filename_{}'.format(i) for i in range(self.patch_level + 1)]

        if inplace:
            self._retrieve('image.processimages',
                           copyvars=column_names,
                           casout=dict(replace=True, blocksize=blocksize,
                                       **self.to_outtable_params()),
                           imagefunctions=[
                               dict(functionoptions=dict(functiontype='RESIZE',
                                                         w=width, h=height))])
            # self._retrieve('table.partition',
            #                casout=dict(replace=True, blocksize=blocksize,
            #                           **self.to_outtable_params())

        else:
            out = self.copy_table()
            out.resize(width=width, height=height)
            return out

    def as_patches(self, x=0, y=0, width=None, height=None, step_size=None,
                   output_width=None, output_height=None, inplace=True):
        '''
        Generate patches from the images in the ImageTable

        Parameters
        ----------
        x : int, optional
            Specify the x location of the top-left corner of the
            first patches.
        y : int, optional
            Specify the y location of the top-left corner of the
            first patches.
        width : int, optional
            Specify the width of the patches.
        height : int, optional
            Specify the width of the patches.
            If not specified, height will be set to be equal to width.
        step_size : int, optional
            Specify the step size of the moving windows for extracting
            the patches.
            Default : None, meaning step_size=width.
        output_width : int, optional
            Specify the output width of the patches.
            If not equal to width, the patches will be resize to the
            output width.
            Default : None, meaning output_width=width.
        output_height : int, optional
            Specify the output height of the patches.
            If not equal to height, the patches will be resize to the
            output height.
            Default : None, meaning output_height=height.
        inplace: boolean, optional
            Specifies whether to update the original table, or create a
            new one.

        Notes
        -----
        By creating crops with fixed window size and moving the window
        along the images.

        Returns
        -------
        ImageTable
            If `inplace=False`
        None
            If `inplace=True`

        '''

        if (width is None) and (height is None):
            width = 224
        if width is None:
            width = height
        if height is None:
            height = width

        if step_size is None:
            step_size = width

        if output_width is None:
            output_width = width
        if output_height is None:
            output_height = height

        blocksize = image_blocksize(output_width, output_height)
        croplist = [dict(sweepimage=True, x=x, y=y,
                         width=width, height=height,
                         stepsize=step_size,
                         outputwidth=output_width,
                         outputheight=output_height)]

        column_names = ['_filename_{}'.format(i) for i in range(self.patch_level + 1)]

        if inplace:
            self._retrieve('image.augmentimages',
                           copyvars=column_names,
                           casout=dict(replace=True, **self.to_outtable_params()),
                           croplist=croplist)

            # The following code generate the latest file name according
            # to the number of patches operations.
            computedvars = '_filename_{}'.format(self.patch_level + 1)
            code = []
            code.append('length _filename_{1} varchar(*);')
            code.append('dot_loc = LENGTH(_filename_{0}) - '
                        'INDEX(REVERSE(_filename_{0}), \'.\')+1;')
            code.append('_filename_{1} = SUBSTR(_filename_{0}, 1, dot_loc-1) || '
                        'compress(\'_\'||x||\'_\'||y||SUBSTR(_filename_{0},dot_loc));')
            code = '\n'.join(code)
            code = code.format(self.patch_level, self.patch_level + 1)

            self._retrieve('table.shuffle',
                           casout=dict(replace=True, blocksize=blocksize,
                                       **self.to_outtable_params()),
                           table=dict(computedvars=computedvars,
                                      computedvarsprogram=code,
                                      **self.to_table_params()))
            self.patch_level += 1

        else:
            out = self.copy_table()
            out.as_patches(x=x, y=y, width=width, height=height, step_size=step_size,
                           output_width=output_width, output_height=output_height)
            return out

    def as_random_patches(self, random_ratio=0.5, x=0, y=0, width=None, height=None,
                          step_size=None, output_width=None, output_height=None,
                          inplace=True):
        '''
        Generate random patches from the images in the ImageTable

        Parameters
        ----------
        random_ratio: double, optional
            Specifies the proportion of the generated patches to output.
        x : int, optional
            Specifies the x location of the top-left corner of the first patches.
        y : int, optional
            Specifies the y location of the top-left corner of the first patches.
        width : int, optional
            Specifies the width of the patches.
        height : int, optional
            Specifies the width of the patches.
            If not specified, height will be set to be equal to width.
        step_size : int, optional
            Specifies the step size of the moving windows for extracting the patches.
            If not specified, it will be set to be equal to width.
        output_width : int, optional
            Specifies the output width of the patches.
            If not specified, it will be set to be equal to width.
        output_height : int, optional
            Specifies the output height of the patches.
            If not specified, it will be set to be equal to height.
        inplace: boolean, optional
            Specifies whether to update the original table, or create a new one.

        Returns
        -------
        :class:`ImageTable`
            If `inplace=True`
        `None`
            If `inplace=False`

        '''
        if (width is None) and (height is None):
            width = 224
        if width is None:
            width = height
        if height is None:
            height = width

        if step_size is None:
            step_size = width

        if output_width is None:
            output_width = width
        if output_height is None:
            output_height = height

        blocksize = image_blocksize(output_width, output_height)

        croplist = [dict(sweepimage=True, x=x, y=y,
                         width=width, height=height,
                         stepsize=step_size,
                         outputwidth=output_width,
                         outputheight=output_height)]

        column_names = ['_filename_{}'.format(i) for i in range(self.patch_level + 1)]

        if inplace:
            self._retrieve('image.augmentimages',
                           copyvars=column_names,
                           casout=dict(replace=True, **self.to_outtable_params()),
                           croplist=croplist,
                           randomratio=random_ratio,
                           writerandomly=True)

            # The following code generate the latest file name according
            # to the number of patches operations.
            computedvars = '_filename_{}'.format(self.patch_level + 1)
            code = []
            code.append('length _filename_{1} varchar(*);')
            code.append('dot_loc = LENGTH(_filename_{0}) - '
                        'INDEX(REVERSE(_filename_{0}),\'.\')+1;')
            code.append('_filename_{1} = SUBSTR(_filename_{0},1,dot_loc-1) || '
                        'compress(\'_\'||x||\'_\'||y||SUBSTR(_filename_{0},dot_loc));')
            code = '\n'.join(code)
            code = code.format(self.patch_level, self.patch_level + 1)

            self._retrieve('table.shuffle',
                           casout=dict(replace=True, blocksize=blocksize,
                                       **self.to_outtable_params()),
                           table=dict(computedvars=computedvars,
                                      computedvarsprogram=code,
                                      **self.to_table_params()))

            self.patch_level += 1

        else:
            out = self.copy_table()
            out.as_random_patches(random_ratio=random_ratio,
                                  x=x, y=y,
                                  width=width, height=height,
                                  step_size=step_size,
                                  output_width=output_width,
                                  output_height=output_height)
            return out

    @property
    def image_summary(self):
        '''
        Summarize the images in the ImageTable

        Returns
        -------
        :class:`pd.Series`

        '''
        out = self._retrieve('image.summarizeimages')['Summary']
        out = out.T.drop(['Column'])[0]
        out.name = None
        return out

    @property
    def label_freq(self):
        '''
        Summarize the distribution of different classes (labels) in the ImageTable

        Returns
        -------
        :class:`pd.Series`

        '''
        out = self._retrieve('simple.freq', table=self, inputs=['_label_'])['Frequency']
        out = out[['FmtVar', 'Level', 'Frequency']]
        out = out.set_index('FmtVar')
        # out.index.name = 'Label'
        out.index.name = None
        out = out.astype('int64')
        return out

    @property
    def channel_means(self):
        '''
        A list of the means of the image intensities in each color channel.

        Returns
        -------
        ( first-channel-mean, second-channel-mean, third-channel-mean )

        '''
        return self.image_summary[['mean1stChannel', 'mean2ndChannel',
                                   'mean3rdChannel']].tolist()

    @property
    def uid(self):
        '''
        A unique ID for each image.

        Returns
        -------


        '''
        file_name = '_filename_{}'.format(self.patch_level)
        uid = self[['_label_', file_name]].to_frame()
        # uid = uid.rename(columns={file_name: '_uid_'})
        return uid



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Residual Network object for deep learning '''

from .layers import (Conv2d, BN, Res, Concat)


class Block(object):
    ''' Block of layers that have special connectivity instead of sequential '''

    def __init__(self):
        self.layers = []
        self.connectivity = {}


class ResBlock(object):
    '''
    Residual block for Residual Network

    Parameters
    ----------
    kernel_size : int, optional.
        Kernel size of the convolution filters.
    n_filters : iter-of-ints, optional.
        List of numbers of filter in each convolution layers.
    strides : iter-of-ints, optional.
        List of stride in each convolution layers.

    Returns
    -------
    :class:`ResBlock`

    '''

    def __init__(self, kernel_sizes=3, n_filters=(16, 16), strides=None):
        if strides is None:
            strides = [1] * len(n_filters)
        elif len(strides) == 1:
            strides = [strides].append([1] * (len(n_filters) - 1))

        if len(strides) != len(n_filters):
            raise ValueError('The length of strides must be equal '
                             'to the length of n_filters.')
        if isinstance(kernel_sizes, int):
            kernel_sizes = [kernel_sizes]
        if len(kernel_sizes) == 1:
            kernel_sizes = [kernel_sizes] * len(n_filters)
        elif len(kernel_sizes) != len(n_filters):
            raise ValueError('The length of kernel_sizes must be equal '
                             'to the length of n_filters.')
        self.config = dict()
        self.config['type'] = 'block'
        self.layers = []

        for n_filter, kernel_size, stride in zip(n_filters, kernel_sizes, strides):
            self.layers.append(Conv2d(n_filters=n_filter,
                                      width=kernel_size,
                                      stride=stride))

        self.layers.append(Res(act='identity'))

    def compile(self, src_layer, block_num):
        '''
        Convert the options into DLPy layer definition.

        Parameters
        ----------
        src_layer : Layer object.
            The source layer for the whole block.
        block_num : int.
            The label of the block. (used to name the layers).

        Returns
        -------
        dict
            A dictionary of keyword-arguments.

        '''
        options = []
        conv_num = 1
        input_layer = src_layer
        for layer in self.layers:
            if layer.config['type'].lower() in ('convo', 'convolution'):
                layer.name = 'R{}C{}'.format(block_num, conv_num)
                conv_num += 1
                layer.src_layers = [input_layer]

            elif layer.config['type'].lower() == 'residual':
                layer.name = 'Res{}'.format(block_num)
                layer.src_layers = [input_layer, src_layer]
            input_layer = layer
            options.append(layer.to_model_params())
        return options


class ResBlockBN(object):
    '''
    Residual block for Residual Network with batch normalization.

    Parameters
    ----------
    kernel_sizes : iter-of-ints, optional.
        Kernel size of the convolution filters.
    n_filters : iter-of-ints, optional.
        List of numbers of filter in each convolution layers.
    strides : iter-of-ints, optional.
        List of stride in each convolution layers.
    batch_norm_first : bool, optional.
        Specify whether to add batch normal layer before conv layer.

    Returns
    -------
    :class:`ResBlockBN`

    '''

    def __init__(self, kernel_sizes=3, n_filters=(16, 16), strides=None,
                 batch_norm_first=True):
        if strides is None:
            strides = [1] * len(n_filters)

        if isinstance(strides, int):
            strides = [strides] + [1] * (len(n_filters) - 1)

        if len(strides) != len(n_filters):
            raise ValueError('The length of strides must be equal '
                             'to the length of n_filters.')
        if isinstance(kernel_sizes, int):
            kernel_sizes = [kernel_sizes] * len(n_filters)

        if len(kernel_sizes) != len(n_filters):
            raise ValueError('The length of kernel_sizes must be equal '
                             'to the length of n_filters.')
        self.config = dict()
        self.config['type'] = 'block'
        self.layers = []
        if batch_norm_first:
            for n_filter, kernel_size, stride in zip(n_filters, kernel_sizes, strides):
                self.layers.append(BN(act='relu'))
                self.layers.append(
                    Conv2d(n_filters=n_filter,
                           width=kernel_size,
                           act='identity',
                           stride=stride,
                           includeBias=False))
        else:
            for n_filter, kernel_size, stride in zip(n_filters, kernel_sizes, strides):
                self.layers.append(
                    Conv2d(n_filters=n_filter,
                           width=kernel_size,
                           act='identity',
                           stride=stride,
                           includeBias=False))
                self.layers.append(BN(act='relu'))

        self.layers.append(Res(act='identity'))

    def compile(self, src_layer, block_num):
        '''
        Convert the options into DLPy layer definition.

        Parameters
        ----------
        src_layer : Layer object.
            The source layer for the whole block.
        block_num : int
            The label of the block. (used to name the layers)

        Returns
        -------
        dict
            A dictionary of keyword-arguments

        '''
        options = []
        conv_num = 1
        bn_num = 1
        input_layer = src_layer
        for layer in self.layers:
            if layer.config['type'].lower() in ('convo', 'convolution'):
                layer.name = 'R{}C{}'.format(block_num, conv_num)
                conv_num += 1
                layer.src_layers = [input_layer]
            elif layer.config['type'].lower() == 'batchnorm':
                layer.name = 'R{}B{}'.format(block_num, bn_num)
                bn_num += 1
                layer.src_layers = [input_layer]
            elif layer.config['type'].lower() == 'residual':
                layer.name = 'Res{}'.format(block_num)
                layer.src_layers = [input_layer, src_layer]
            input_layer = layer
            options.append(layer.to_model_params())
        return options


class ResBlock_Caffe(object):
    '''
    Residual block for Residual Network with batch normalization.

    '''

    def __init__(self, kernel_sizes=3, n_filters=(16, 16), strides=None,
                 batch_norm_first=False, conv_short_cut=False):
        if strides is None:
            strides = [1] * len(n_filters)

        if isinstance(strides, int):
            strides = [strides] + [1] * (len(n_filters) - 1)

        if len(strides) != len(n_filters):
            raise ValueError('The length of strides must be equal '
                             'to the length of n_filters.')
        if isinstance(kernel_sizes, int):
            kernel_sizes = [kernel_sizes] * len(n_filters)

        if len(kernel_sizes) != len(n_filters):
            raise ValueError('The length of kernel_sizes must be equal '
                             'to the length of n_filters.')
        self.config = dict()
        self.config['type'] = 'block'
        self.layers = []
        self.conv_short_cut = conv_short_cut
        if batch_norm_first:
            if conv_short_cut:
                self.layers.append(BN(act='relu'))
                self.layers.append(
                    Conv2d(n_filters=n_filters[-1],
                           width=1,
                           act='identity',
                           stride=strides[0],
                           includeBias=False))
            self.layers.append(Res(act='identity'))

            for n_filter, kernel_size, stride in zip(n_filters, kernel_sizes, strides):
                self.layers.append(BN(act='relu'))
                self.layers.append(
                    Conv2d(n_filters=n_filter,
                           width=kernel_size,
                           act='identity',
                           stride=stride,
                           includeBias=False))
        else:
            if conv_short_cut:
                self.layers.append(
                    Conv2d(n_filters=n_filters[-1],
                           width=1,
                           act='identity',
                           stride=strides[0],
                           includeBias=False))
                self.layers.append(BN(act='identity'))

            for n_filter, kernel_size, stride, act in zip(n_filters,
                                                          kernel_sizes,
                                                          strides,
                                                          ['relu', 'relu', 'identity']):
                self.layers.append(
                    Conv2d(n_filters=n_filter,
                           width=kernel_size,
                           act='identity',
                           stride=stride,
                           includeBias=False))
                self.layers.append(BN(act=act))

            self.layers.append(Res(act='relu'))

    def compile(self, src_layer, block_num):
        '''
        Parameters
        ----------
        src_layer : Layer object.
            The source layer for the whole block.
        block_num : int
            The label of the block. (used to name the layers)

        Returns
        -------
        dict
            A dictionary of keyword-arguments

        '''
        options = []
        conv_num = 1
        bn_num = 1
        input_layer = src_layer
        if self.conv_short_cut:
            for layer in self.layers[:2]:
                if layer.config['type'].lower() in ('convo', 'convolution'):
                    layer.name = 'R{}C{}'.format(block_num, 0)
                    conv_num += 1
                    layer.src_layers = [input_layer]
                elif layer.config['type'].lower() == 'batchnorm':
                    layer.name = 'R{}B{}'.format(block_num, 0)
                    bn_num += 1
                    layer.src_layers = [input_layer]
                input_layer = layer
                options.append(layer.to_model_params())
            short_cut_layer = layer
            input_layer = src_layer
            for layer in self.layers[2:]:
                if layer.config['type'].lower() in ('convo', 'convolution'):
                    layer.name = 'R{}C{}'.format(block_num, conv_num)
                    conv_num += 1
                    layer.src_layers = [input_layer]
                elif layer.config['type'].lower() == 'batchnorm':
                    layer.name = 'R{}B{}'.format(block_num, bn_num)
                    bn_num += 1
                    layer.src_layers = [input_layer]
                elif layer.config['type'].lower() == 'residual':
                    layer.name = 'Res{}'.format(block_num)
                    layer.src_layers = [input_layer, short_cut_layer]
                input_layer = layer
                options.append(layer.to_model_params())
        else:
            for layer in self.layers:
                if layer.config['type'].lower() in ('convo', 'convolution'):
                    layer.name = 'R{}C{}'.format(block_num, conv_num)
                    conv_num += 1
                    layer.src_layers = [input_layer]
                elif layer.config['type'].lower() == 'batchnorm':
                    layer.name = 'R{}B{}'.format(block_num, bn_num)
                    bn_num += 1
                    layer.src_layers = [input_layer]
                elif layer.config['type'].lower() == 'residual':
                    layer.name = 'Res{}'.format(block_num)
                    layer.src_layers = [input_layer, src_layer]
                input_layer = layer
                options.append(layer.to_model_params())
        return options


class DenseNetBlock(object):
    '''
    DenseNet block

    Parameters
    ----------
    n_cells : int
        Number of cells
    kernel_size : int
        Size of the kernel
    n_filter : int
        Number of filters
    stride : int
        Size of the stride

    Returns
    -------
    :class:`DenseNetBlock`

    '''

    def __init__(self, n_cells=4, kernel_size=3, n_filter=12, stride=1):
        self.config = dict()
        self.config['type'] = 'block'
        self.layers = []
        for _ in range(n_cells):
            self.layers.append(BN(act='relu'))
            self.layers.append(
                Conv2d(n_filters=n_filter,
                       width=kernel_size,
                       act='relu',
                       stride=stride,
                       includeBias=False))
            self.layers.append(Concat(act='identity'))

    def compile(self, src_layer, block_num):
        '''
        Convert the options into DLPy layer definition.

        Parameters
        ----------
        src_layer : Layer object.
            The source layer for the whole block.
        block_num : int
            The label of the block. (used to name the layers)

        Returns
        -------
        dict
            A dictionary of keyword-arguments

        '''
        options = []
        conv_num = 1
        bn_num = 1
        concat_num = 1
        input_layer = src_layer
        for layer in self.layers:
            if layer.config['type'].lower() in ('convo', 'convolution'):
                layer.name = 'D{}C{}'.format(block_num, conv_num)
                conv_num += 1
                layer.src_layers = [input_layer]
            elif layer.config['type'].lower() == 'batchnorm':
                layer.name = 'D{}B{}'.format(block_num, bn_num)
                bn_num += 1
                layer.src_layers = [input_layer]
            elif layer.config['type'].lower() == 'concat':
                layer.name = 'D{}Concat{}'.format(block_num, concat_num)
                concat_num += 1
                layer.src_layers = [input_layer, src_layer]
                src_layer = layer
            input_layer = layer
            options.append(layer.to_model_params())
        return options



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Convert keras model to sas models '''

import os
import sys

from keras import backend as K

from .write_keras_model_parm import write_keras_hdf5
from .write_sas_code import (write_input_layer, write_convolution_layer,
                             write_batch_norm_layer, write_pooling_layer,
                             write_residual_layer, write_full_connect_layer,
                             write_main_entry)

computation_layer_classes = ['averagepooling2d', 'maxpooling2d', 'conv2d',
                             'dense', 'batchnormalization', 'add']
dropout_layer_classes = ['averagepooling2d', 'maxpooling2d', 'conv2d', 'dense']


class KerasParseError(ValueError):
    '''
    Used to indicate an error in parsing Keras model definition

    '''


def keras_to_sas(model, model_name=None):
    output_code = ''
    layer_activation = {}
    src_layer = {}
    layer_dropout = {}
    if model_name is None:
        model_name = model.name
    for layer in model.layers:
        class_name = layer.__class__.__name__.lower()
        if (class_name in computation_layer_classes):
            comp_layer_name = find_previous_computation_layer(
                model, layer.name, computation_layer_classes)
            source_str = make_source_str(comp_layer_name)
            src_layer.update({layer.name: source_str})
        elif (class_name == 'activation'):
            tmp_name = find_previous_computation_layer(
                model, layer.name, computation_layer_classes)
            tmp_act = extract_activation(layer)
            layer_activation.update({tmp_name[0]: tmp_act})
        elif (class_name == 'dropout'):
            tmp = find_next_computation_layer(model, layer, dropout_layer_classes)
            dconfig = layer.get_config()
            layer_dropout.update({tmp: dconfig['rate']})

    # if first layer is not an input layer, generate the correct
    # input layer code for a SAS deep learning model
    layer = model.layers[0]
    if (layer.__class__.__name__.lower() != 'inputlayer'):
        sas_code = keras_input_layer(layer, model_name, False)
        # write SAS code for input layer
        if sas_code:
            output_code = output_code + sas_code + '\n\n'
        else:
            raise KerasParseError('Unable to generate an input layer')

    # extract layers and apply activation functions as needed
    for layer in model.layers:
        class_name = layer.__class__.__name__.lower()

        sas_code = None

        # determine activation function
        if (class_name in ['conv2d', 'batchnormalization', 'add', 'dense']):
            if (layer.name in layer_activation.keys()):
                act_func = layer_activation[layer.name]
            else:
                act_func = None
        else:
            act_func = None

        # average/max pooling
        if (class_name in ['averagepooling2d', 'maxpooling2d']):
            sas_code = keras_pooling_layer(layer, model_name, class_name,
                                           src_layer, layer_dropout)
        # 2D convolution
        elif (class_name == 'conv2d'):
            sas_code = keras_convolution_layer(layer, model_name, act_func,
                                               src_layer, layer_dropout)
        # batch normalization
        elif (class_name == 'batchnormalization'):
            sas_code = keras_batchnormalization_layer(layer, model_name,
                                                      act_func, src_layer)
        # input layer
        elif (class_name == 'inputlayer'):
            sas_code = keras_input_layer(layer, model_name, True)
        # add
        elif (class_name == 'add'):
            sas_code = keras_residual_layer(layer, model_name,
                                            act_func, src_layer)
        elif (class_name in ['activation', 'flatten', 'dropout']):
            pass
        # fully connected
        elif (class_name == 'dense'):
            sas_code = keras_full_connect_layer(layer, model_name, act_func,
                                                src_layer, layer_dropout)
        else:
            print('WARNING: ' + class_name + ' is an unsupported layer '
                                             'type - your SAS model may be incomplete')

        # write SAS code associated with Keras layer
        if sas_code:
            output_code = output_code + sas_code + '\n\n'
        elif (class_name not in ['activation', 'flatten', 'dropout']):
            print('WARNING: unable to generate SAS definition '
                  'for layer ' + class_name)
    return output_code


# create SAS pooling layer
def keras_pooling_layer(layer, model_name, class_name, src_layer, layer_dropout):
    '''
    Extract pooling layer parameters from layer definition object

    Parameters
    ----------
    layer : Layer object
       Pooling layer
    model_name : string
       Deep learning model name
    class_name : string
       Layer class
    src_layer : list-of-Layer objects
       Layer objects corresponding to source layer(s) for
       pooling layer
    layer_dropout : dict
       Dictionary containing dropout layer names (keys) and dropout rates (values)

    Returns
    -------
    string
        String value with SAS deep learning pooling layer definition

    '''
    config = layer.get_config()
    strides = config['strides']
    pool_size = config['pool_size']

    # pooling size
    width, height = pool_size

    # stride
    if (strides[0] == strides[1]):
        step = strides[0]
    else:
        raise KerasParseError('Unequal strides in vertical/horizontal '
                              'directions for pooling layer')

    # pooling type
    if (class_name == 'averagepooling2d'):
        type = 'mean'
    elif (class_name == 'maxpooling2d'):
        type = 'max'
    else:
        raise KerasParseError('Pooling type ' + class_name +
                              ' is not supported yet')

    # extract source layer(s)
    if (layer.name in src_layer.keys()):
        source_str = src_layer[layer.name]
    else:
        raise KerasParseError('Unable to determine source layer for '
                              'pooling layer = ' + layer.name)

    # set dropout
    if (layer.name in layer_dropout.keys()):
        dropout = layer_dropout[layer.name]
    else:
        dropout = 0.0

    return write_pooling_layer(model_name=model_name, layer_name=layer.name,
                               width=str(width), height=str(height), stride=str(step),
                               type=type, dropout=str(dropout), src_layer=source_str)


# create SAS 2D convolution layer
def keras_convolution_layer(layer, model_name, act_func, src_layer, layer_dropout):
    '''
    Extract convolution layer parameters from layer definition object

    Parameters
    ----------
    layer : Layer object
       Convolution layer
    model_name : string
       Deep learning model name
    act_func : string
       Keras activation function
    src_layer : list-of-Layer objects
       Layer objects corresponding to source layer(s) for
       convolution layer
    layer_dropout : dict
       Dictionary containing dropout layer names (keys) and dropout rates (values)

    Returns
    -------
    string
        String value with SAS deep learning convolution layer definition

    '''
    config = layer.get_config()

    strides = config['strides']
    kernel_size = config['kernel_size']

    # activation
    if (act_func):
        layer_act_func = map_keras_activation(layer, act_func)
    elif ('activation' in config.keys()):
        layer_act_func = map_keras_activation(layer, config['activation'])
    else:
        layer_act_func = 'identity'

    # kernel size
    width, height = kernel_size

    # stride
    if (strides[0] == strides[1]):
        step = strides[0]
    else:
        raise KerasParseError('Unequal strides in vertical/horizontal '
                              'directions for convolution layer')

    # bias term
    if (config['use_bias']):
        bias_str = 'False'
    else:
        bias_str = 'True'

    # number of filters
    nrof_filters = config['filters']

    # extract source layer(s)
    if (layer.name in src_layer.keys()):
        source_str = src_layer[layer.name]
    else:
        raise KerasParseError('Unable to determine source layer for '
                              'convolution layer = ' + layer.name)

    # set dropout
    if (layer.name in layer_dropout.keys()):
        dropout = layer_dropout[layer.name]
    else:
        dropout = 0.0

    return write_convolution_layer(model_name=model_name, layer_name=layer.name,
                                   nfilters=str(nrof_filters), width=str(width),
                                   height=str(height), stride=str(step),
                                   nobias=bias_str, activation=layer_act_func,
                                   dropout=str(dropout), src_layer=source_str)


# create SAS batch normalization layer
def keras_batchnormalization_layer(layer, model_name, act_func, src_layer):
    '''
    Extract batch normalization layer parameters from layer definition object

    Parameters
    ----------
    layer : Layer object
       Batch nornalization layer
    model_name : string
       Deep learning model name
    act_func : string
       Keras activation function
    src_layer : list-of-Layer objects
       Layer objects corresponding to source layer(s) for
       batch normalization layer

    Returns
    -------
    string
        String value with SAS deep learning batch normalization layer definition

    '''
    config = layer.get_config()

    # extract source layer(s)
    if (layer.name in src_layer.keys()):
        source_str = src_layer[layer.name]
    else:
        raise KerasParseError('Unable to determine source layer for '
                              'batch normalization layer = ' + layer.name)

    # activation
    if (act_func):
        layer_act_func = map_keras_activation(layer, act_func)
    elif ('activation' in config.keys()):
        layer_act_func = map_keras_activation(layer, config['activation'])
    else:
        layer_act_func = 'identity'

    return write_batch_norm_layer(model_name=model_name, layer_name=layer.name,
                                  activation=layer_act_func, src_layer=source_str)


# create SAS input layer
def keras_input_layer(layer, model_name, input_layer):
    '''
    Extract input layer parameters from layer definition object

    Parameters
    ----------
    layer : Layer object
       Input layer
    model_name : string
       Deep learning model name
    input_layer : boolean
       Indicate whether layer name given (True) or not (False)

    Returns
    -------
    string
        String value with SAS deep learning input layer definition

    '''
    config = layer.get_config()

    if (K.image_data_format() == 'channels_first'):
        dummy, C, H, W = config['batch_input_shape']
    else:
        dummy, H, W, C = config['batch_input_shape']

    # generate name based on whether layer is actually an input layer
    if (input_layer):
        input_name = config['name']
    else:
        input_name = config['name'] + '_input'

    # TBD: fix scale, now default value for scale
    scale = 1.0

    return write_input_layer(model_name=model_name, layer_name=input_name,
                             channels=str(C), width=str(W),
                             height=str(H), scale=str(scale))


# create SAS residual layer
def keras_residual_layer(layer, model_name, act_func, src_layer):
    '''
    Extract residual layer parameters from layer definition object

    Parameters
    ----------
    layer : Layer object
       Add layer
    model_name : string
       Deep learning model name
    act_func : string
       Keras activation function
    src_layer : list-of-Layer objects
       Layer objects corresponding to source layer(s) for
       residual layer

    Returns
    -------
    string
        String value with SAS deep learning residual layer definition

    '''
    config = layer.get_config()

    # extract source layer(s)
    if (layer.name in src_layer.keys()):
        source_str = src_layer[layer.name]
    else:
        raise KerasParseError('Unable to determine source layers for '
                              'residual layer = ' + layer.name)

    # activation
    if (act_func):
        layer_act_func = map_keras_activation(layer, act_func)
    elif ('activation' in config.keys()):
        layer_act_func = map_keras_activation(layer, config['activation'])
    else:
        layer_act_func = 'identity'

    return write_residual_layer(model_name=model_name, layer_name=layer.name,
                                activation=layer_act_func, src_layer=source_str)


# create SAS fully connected layer
def keras_full_connect_layer(layer, model_name, act_func, src_layer, layer_dropout):
    '''
    Extract fully connected layer parameters from layer definition object

    Parameters
    ----------
    layer : Layer object
       Fully connected layer
    model_name : string
       Deep learning model name
    act_func : string
       Keras activation function
    src_layer : list-of-Layer objects
       Layer objects corresponding to source layer(s) for
       fully connected layer
    layer_dropout : dict
       Dictionary containing dropout layer names (keys) and dropout rates (values)

    Returns
    -------
    string
        String value with SAS deep learning fully connected layer definition

    '''
    config = layer.get_config()

    # activation
    if (act_func):
        layer_act_func = map_keras_activation(layer, act_func)
    elif ('activation' in config.keys()):
        layer_act_func = map_keras_activation(layer, config['activation'])
    else:
        layer_act_func = 'identity'

    # check whether SOFTMAX or other type of activation
    if (layer_act_func.lower() == 'softmax'):
        layer_type = 'output'
    else:
        layer_type = 'fullconnect'

    # number of neurons
    nrof_neurons = config['units']

    # bias term
    if (config['use_bias']):
        bias_str = 'False'
    else:
        bias_str = 'True'

        # set dropout
    if (layer.name in layer_dropout.keys()):
        dropout = layer_dropout[layer.name]
    else:
        dropout = 0.0

    # extract source layer(s)
    if (layer.name in src_layer.keys()):
        source_str = src_layer[layer.name]
    else:
        raise KerasParseError('Unable to determine source layer for '
                              'fully connected layer = ' + layer.name)

    return write_full_connect_layer(model_name=model_name, layer_name=layer.name,
                                    nrof_neurons=str(nrof_neurons), nobias=bias_str,
                                    activation=layer_act_func, type=layer_type,
                                    dropout=str(dropout), src_layer=source_str)


def map_keras_activation(layer, act_func):
    '''
    Map Keras activation function(s) to SAS activation function(s)

    Parameters
    ----------
    layer : Layer object
       Current layer definition
    act_func : string
       Keras activation function

    Returns
    -------
    string
        SAS activation type

    '''
    class_name = layer.__class__.__name__.lower()
    # convolution layer
    if (class_name in ['conv2d', 'batchnormalization']):
        map_dict = {'softmax': None, 'elu': 'elu', 'selu': None,
                    'softplus': 'softplus', 'softsign': None,
                    'relu': 'relu', 'tanh': 'tanh', 'sigmoid': 'sigmoid',
                    'hard_sigmoid': None, 'linear': 'identity'}
    elif (class_name == 'dense'):
        map_dict = {'softmax': 'softmax', 'elu': 'elu', 'selu': None,
                    'softplus': 'softplus', 'softsign': None,
                    'relu': 'relu', 'tanh': 'tanh', 'sigmoid': 'sigmoid',
                    'hard_sigmoid': None, 'linear': 'identity'}
    elif (class_name == 'add'):
        map_dict = {'softmax': None, 'elu': None, 'selu': None,
                    'softplus': None, 'softsign': None, 'relu': 'relu',
                    'tanh': None, 'sigmoid': None, 'hard_sigmoid': None,
                    'linear': 'identity'}
    else:
        raise KerasParseError('SAS does not support activation functions '
                              'for layer ' + layer.name)

    if (act_func.lower() in map_dict.keys()):
        sas_act_func = map_dict[act_func.lower()]
        if not sas_act_func:
            raise KerasParseError('Activation function ' + act_func + ' not supported')
    else:
        raise KerasParseError('Unknown Keras activation function = ' + act_func)

    return sas_act_func


# extract parameters from activation layer
def extract_activation(layer):
    '''
    Extract activation from layer definition object

    Parameters
    ----------
    layer : Layer object
       Activation layer

    Returns
    -------
    string
        String value with Keras activation function

    '''
    config = layer.get_config()

    return config['activation']


def find_next_computation_layer(model, layer, computation_layer_list):
    '''
    Extract the name of the computation layer following the current layer

    Parameters
    ----------
    model : Model object
       Keras deep learning model
    layer : Layer object
       Current layer object
    computation_layer_list : list
       List of computation layers supported by SAS

    Returns
    -------
    string
        String value with name of next computation layer

    '''

    if (len(layer._outbound_nodes) > 1):
        raise KerasParseError('Unable to determine next computation layer '
                              'for layer = ' + layer.name)
    else:
        node_config = layer._outbound_nodes[0].get_config()
        layer_name = node_config['outbound_layer']
        tmp_layer = model.get_layer(name=layer_name)
        while (tmp_layer.__class__.__name__.lower() not in computation_layer_list):
            if (len(tmp_layer._outbound_nodes) > 1):
                raise KerasParseError('Unable to determine next computation layer '
                                      'for layer = ' + layer.name)
                break
            else:
                node_config = tmp_layer._outbound_nodes[0].get_config()
                layer_name = node_config['outbound_layer']
                tmp_layer = model.get_layer(name=layer_name)

    return layer_name


def find_previous_computation_layer(model, layer_name, computation_layer_list):
    '''
    Extract the name of the computation layer prior to the current layer

    Parameters
    ----------
    model : Model object
       Keras deep learning model
    layer_name : string
       Current layer name
    computation_layer_list : list
       List of computation layers supported by SAS

    Returns
    -------
    string
        String value with name of previous computation layer

    '''

    layer = model.get_layer(name=layer_name)
    if (len(layer._inbound_nodes) > 1):
        raise KerasParseError('Unable to determine previous computation '
                              'layer(s) for layer = ' + layer_name)
        return None
    else:
        src_layer_name = []
        node_config = layer._inbound_nodes[0].get_config()
        for lname in node_config['inbound_layers']:
            tmp_layer = model.get_layer(name=lname)
            prev_layer = tmp_layer
            while (tmp_layer.__class__.__name__.lower() not in computation_layer_list):
                # check for root node
                node_config = tmp_layer._inbound_nodes[0].get_config()
                if (len(node_config['inbound_layers']) > 1):
                    raise KerasParseError('Unable to determine previous computation '
                                          'layer(s) for layer = ' + layer_name)
                    return None
                elif (len(node_config['inbound_layers']) == 1):
                    tmp_name = node_config['inbound_layers'][0]
                    prev_layer = tmp_layer
                    tmp_layer = model.get_layer(name=tmp_name)
                else:
                    tmp_layer = prev_layer
                    break

            src_layer_name.append(tmp_layer.name)

        return src_layer_name


def make_source_str(layer_name):
    '''
    Create a string value representing a Python list of source layers

    Parameters
    ----------
    layer_name : string
       List of Python layer names

    Returns
    -------
    string
        String representation of list of Python layer names

    '''
    source_str = []
    for ii in range(len(layer_name)):
        source_str.append(layer_name[ii])
    return repr(source_str)
